/********************************************************************/
/*                                                                  */
/*   MQTIMES2 supports the following parameters                     */
/*                                                                  */
/*      -c message count (number to read before ending)             */
/*      -t maximum timeout value (end if no messages in this time)  */
/*      -q the name of the input queue (required)                   */
/*      -m queue manager name (optional)                            */
/*      -d drain queue before starting measurements                 */
/*      -b batch size (number of messages in a unit of work)        */
/*      -r report interval (seconds between report lines)           */
/*      -p name of file containing PAN reply data                   */
/*      -n name of file containing NAN reply data                   */
/*                                                                  */
/*    if no queue manager is specified, the default queue manager   */
/*    is used.                                                      */
/*                                                                  */
/********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* includes for MQI */
#include <cmqc.h>

#ifdef SOLARIS
#include <ctype.h>
#include <unistd.h>
#endif

#define MAX_MESSAGE_LENGTH	64 * 65536 + 1
#define TIME_OUT_DEF		120
#define DEF_SYNC			25
#define MAX_SYNC_ALLOW		1000

/* global error switch */
	int		err=0;

/* Globally declared variables (from arguments) */
	int		maxmsgs=0;
	int		maxtime=TIME_OUT_DEF;
	int		reportInterval=1;
	int		drainQ=0;
	int		batchSize=DEF_SYNC;
	int		fileSizePAN=0;
	int		fileSizeNAN=0;
	char	*fileDataPAN=NULL;
	char	*fileDataNAN=NULL;
	char	qname[MQ_Q_NAME_LENGTH + 1];
	char	qmname[MQ_Q_MGR_NAME_LENGTH + 1];

static char copyright[] = "(C) Copyright IBM Corp, 2001/2002";
static char Version[]=\
"@(#)MQTimes2 V1.2 - MQSI V2 Performance results tool  - Jim MacNair ";

#ifdef _DEBUG
static char Level[]="mqtimes2.c V1.2 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="mqtimes2.c V1.2 Release version ("__DATE__" "__TIME__")";
#endif

void checkerror(const char *mqcalltype, MQLONG compcode, MQLONG reason, char *resource)

{
	if ((compcode != MQCC_OK) || (reason != MQRC_NONE))
	{
		printf("MQSeries error with %s on %s - compcode = %d, reason = %d\n",
				mqcalltype, resource, compcode, reason);
	}
}

void captureString(char * value, char * arg, const int maxlen)

{
	if ((int) strlen(arg) < maxlen)
	{
		strcpy(value, arg);
	}
	else
	{
		/* issue an error message */
		printf("*****Value of constant exceeds maximum length\n");

		/* Set the return code based on the message number */
		err = 94;
	}
}

int getArgPointer(int argc,
				  char *argv[],
				  int index,
				  char ** parmData,
				  const char * parmDesc)

{
	if (strlen(argv[index]) > 2)
	{
		/* point to the argument data */
		(*parmData) = argv[index] + 2;
	}
	else
	{
		index++;

		if (index < argc)
		{
			/* point to the argument data */
			(*parmData) = argv[index];
		}
		else
		{
			/* issue an error message */
			printf("*****%s parameter is missing value\n", parmDesc);
			
			/* Set the return code to indicate the error */
			err = 95;

			/*point to something valid */
			(*parmData) = argv[index - 1] + 2;
		}
	}


	return index;
}

int processFileArg(int argc,
				   char *argv[], 
				   int index, 
				   int * fileSize, 
				   char **fileData,
				   const char * parmDesc)

{
	FILE	*inputFile;
	char	*fileName;

	index = getArgPointer(argc, argv, index, &fileName, parmDesc);

	if (index < argc)
	{
		/* process the file */
		/* look for a file name that contains the PAN data */
		inputFile = fopen(fileName, "rb");
		if (NULL == inputFile)
		{
			/* error opening PAN file */
			printf("***** unable to open PAN file data for %s\n", argv[5]);
		}
		else
		{
			/* read the PAN file data */
			/* get the length of the data */
			fseek(inputFile, 0L, SEEK_END);
			(*fileSize) = ftell(inputFile);

			/* read the data from the beginning of the file */
			fseek(inputFile, 0L, SEEK_SET);
			(*fileData) = (char *)malloc((*fileSize) + 1);

			if (fileSize > 0)
			{
				fread((*fileData), 1, (*fileSize), inputFile);
			}

			(*fileData)[(*fileSize)] = 0;

			/* close the file */
			fclose(inputFile);
		}
	}

	return index;
}

int processNumArg(int argc,
				  char *argv[], 
				  int index, 
				  int * value, 
				  const char * parmDesc)

{
	char *argData;

	index = getArgPointer(argc, argv, index, &argData, parmDesc);

	if (index < argc)
	{
		/* convert to an integer */
		(*value) = atoi(argData);
	}

	return index;
}

int processIndArg(int argc, 
				  char *argv[], 
				  int index, 
				  char * parm, 
				  int parmsize, 
				  const char * parmDesc)

{
	char *argData;

	index = getArgPointer(argc, argv, index, &argData, parmDesc);

	if (index < argc)
	{
		/* value is included with parameter */
		captureString(parm, argData, parmsize);
	}

	return index;
}

void processArgs(int argc, char **argv)

{
	int		i;
	int		foundit;
	char	ch;

	i = 1;
	while ((i < argc) && (0 == err))
	{
		foundit = 0;

		if ('-' == argv[i][0])
		{
			ch = toupper(argv[i][1]);

			/* check for option */
			if ((0 == foundit) && ('Q' == ch))
			{
				foundit = 1;

				i = processIndArg(argc, 
								  argv, 
								  i,
								  qname, 
								  sizeof(qname), 
								  "queue name (-q)");
			}

			/* check for option */
			if ((0 == foundit) && ('M' == ch))
			{
				foundit = 1;

				i = processIndArg(argc, 
								  argv, 
								  i,
								  qmname, 
								  sizeof(qmname), 
								  "queue manager name (-m)");
			}

			/* check for option */
			if ((0 == foundit) && ('T' == ch))
			{
				foundit = 1;

				i = processNumArg(argc,
								  argv, 
								  i,
								  &maxtime, 
								  "Timeout (-t)");

				if (maxtime < 30)
				{
					printf("Timeout parameter (%d) invalid or below minimum of 30 seconds\n", maxtime);
					maxtime = TIME_OUT_DEF;
					printf("Set to default value of %d\n", maxtime);
				}
			}

			/* check for option */
			if ((0 == foundit) && ('B' == ch))
			{
				foundit = 1;

				i = processNumArg(argc,
								  argv, 
								  i,
								  &batchSize, 
								  "Messages within unit of work (-b)");

				if ((batchSize < 1) || (batchSize > MAX_SYNC_ALLOW))
				{
					printf("Invalid sync point interval %s\n", batchSize);
					batchSize = DEF_SYNC;
				}

				printf("Sync point interval set to %d\n", batchSize);
			}

			/* check for option */
			if ((0 == foundit) && ('C' == ch))
			{
				foundit = 1;

				i = processNumArg(argc,
								  argv, 
								  i,
								  &maxmsgs, 
								  "Message count (-c)");
			}

			/* check for option */
			if ((0 == foundit) && ('R' == ch))
			{
				foundit = 1;

				i = processNumArg(argc,
								  argv, 
								  i,
								  &reportInterval, 
								  "Report interval (-r)");

				if (reportInterval > 100)
				{
					printf("reportInterval cannot exceed 100 - reset to 100\n");
					reportInterval = 100;
				}
			}

			/* check for option */
			if ((0 == foundit) && ('D' == ch))
			{
				foundit = 1;
				drainQ = 1;
			}

			/* check for option */
			if ((0 == foundit) && ('P' == ch))
			{
				foundit = 1;

				i = processFileArg(argc,
								   argv, 
								   i,
								   &fileSizePAN, 
								   &fileDataPAN,
								  "PAN Message data (-p)");

				printf("Positive acknowledgement report messages will be sent (data length %d)\n", 
					   fileSizePAN);
			}

			/* check for option */
			if ((0 == foundit) && ('N' == ch))
			{
				foundit = 1;

				i = processFileArg(argc,
								   argv, 
								   i,
								   &fileSizeNAN, 
								   &fileDataNAN,
								  "NAN Message data (-p)");

				printf("Negative acknowledgement report messages will be sent (data length %d)\n", 
					   fileSizeNAN);
			}
		}

		/* did we recognize the parameter? */
		if (0 == foundit)
		{
			/* issue an error message */
			printf("*****Unrecognized command line parameter found %c\n", argv[i][0]);

			/* Set the return code based on the message number */
			err = 93;
		}

		i++;
	}

	if (maxmsgs <= 0)
	{
		printf("Message count parameter must be > 0");
		err = 1;
	}

	if (0 == qname[0])
	{
		printf("Queue name parameter is required\n");
		err = 1;
	}
}

void printHelp(char *pgmName)

{
	printf("%s\n", Level);
	printf("format is:\n");
	printf("   %s -c msgCount <-t timeout> -q queue <-m queue manager> <-d> <-b nnn> <-r nnn> <-p PAN reply data file> <-n NAN reply data file> \n\n", pgmName);
	printf("    msgCount is the number of messages to read before stopping.\n");
	printf("    timeout is the maximum time in seconds to wait for a\n");
	printf("    message before stopping without reading the maximum\n");
	printf("    number of messages.  queue is the name of the queue\n");
	printf("    to read messages from.  queue manager is the name of the\n");
	printf("    queue manager that holds the input queue.  If the program must\n");
	printf("    respond to either PAN or NAN report options, a file containing\n");
	printf("    the data to be used for the reply message must be provided.\n");
	printf("    The -d option will drain the queue before starting the measurement.\n");
	printf("    Any messages in the queue will be discarded. The -b option allows\n");
	printf("    the number of messages read in a single unit of work to be specified.\n");
	printf("    The -r option allows the report interval to be specified.  It can be\n");
	printf("    from 1 to 20.\n");
}

int main(int argc, char **argv)
{
	int		i;
	int		msgcount=0;
	int		totcount=0;
	int		maxrate=0;
	int		firstsec=0;
	int		secondcount=0;
	int		totalbytes=0;
	int		reportCount=0;
	int		avgbytes;
	int		uow=0;
	int		drainCount=0;
	int		recent10;
	int		recent10sec;
	int		last10[10];			/* Average rate of last 10 intervals */
	int		last10secs[10];		/* time of last 10 intervals */
	int		time1sec;
	int		time2sec;
	int		hh;
	int		mm;
	int		ss;
	MQLONG	report;				/* MQ report options */
	double	avgrate;
	double	fCount;
	MQHCONN	qm=0;
	MQHOBJ	q=0;
	MQHOBJ	replyq=0;
	MQLONG	compcode;
	MQLONG	reason;
	MQLONG	datalen=0;
	MQOD	objdesc = {MQOD_DEFAULT};
	MQMD	msgdesc = {MQMD_DEFAULT};
	MQOD	replyObjdesc = {MQOD_DEFAULT};
	MQMD	replyMsgdesc = {MQMD_DEFAULT};
	MQLONG	openopt = 0;
	MQGMO	mqgmo = {MQGMO_DEFAULT};
	MQPMO	mqpmo = {MQPMO_DEFAULT};
	MQMD	replyMsg = {MQMD_DEFAULT};
	char	replyQname[MQ_Q_NAME_LENGTH + 1];
	char	replyQMname[MQ_Q_MGR_NAME_LENGTH + 1];
	char	*msgdata;
	char	tempnumb[32];
	char	prevtime[9];
	char	currtime[9];
	MQLONG	prevtime_secs;
	MQLONG  currtime_secs;
	char	*msgPtr;
	char	msgArea[10240];

	printf("%s program start\n", Level);

	/* print the copyright statement */
	printf("Copyright (c) IBM Corp., 2001/2002\n");

	/* allocate a buffer for the message */
	msgdata = (char *)malloc(MAX_MESSAGE_LENGTH);

	/* initialize the work areas */
	memset(qmname, 0, sizeof(qmname));
	memset(qname, 0, sizeof(qname));
	memset(replyQMname, 0, sizeof(replyQMname));
	memset(replyQname, 0, sizeof(replyQname));
	memset(msgdata, 0, sizeof(msgdata));
	memset(prevtime, 0, sizeof(prevtime));
	memset(currtime, 0, sizeof(currtime));
	memset(msgArea, 0, sizeof(msgArea));

	msgPtr = msgArea;

	for (i=0; i<10; i++)
	{
		last10[i] = 0;
		last10secs[i] = 0;
	}

	fileDataPAN = NULL;
	fileDataNAN = NULL;

	prevtime_secs = currtime_secs = 0;

	/* check for too few input parameters */
	if (argc < 4)
	{
		printHelp(argv[0]);
		exit(99);
	}

	/* check for help request */
	if ((argv[1][0] == '?') || (argv[1][1] == '?'))
	{
		printHelp(argv[0]);
		exit(0);
	}

	/* process any command line arguments */
	processArgs(argc, argv);

	if (err != 0)
	{
		printHelp(argv[0]);
		exit(99);
	}

	/* Connect to the queue manager */
	printf("connecting to queue manager %s\n",qmname);
	MQCONN(qmname, &qm, &compcode, &reason);

	/* check for errors */
	checkerror((char *)"MQCONN", compcode, reason, qmname);
	if (compcode != MQCC_OK)
	{
		return 98;
	}

	/* set the queue open options */
	strncpy(objdesc.ObjectName, qname, MQ_Q_NAME_LENGTH);
	openopt = MQOO_INPUT_SHARED | MQOO_FAIL_IF_QUIESCING;

	/* open the queue for input */
	printf("opening queue %s for input\n", qname);
	MQOPEN(qm, &objdesc, openopt, &q, &compcode, &reason);

	/* check for errors */
	checkerror((char *)"MQOPEN", compcode, reason, qname);
	if (compcode != MQCC_OK)
	{
		return 97;
	}

	/* check to see if queue is to be drained before run starts */
	if (drainQ > 0)
	{
		printf("draining queue\n");
		uow = 0;
		compcode = MQCC_OK;
		while (compcode == MQCC_OK)
		{
			mqgmo.Options = MQGMO_NO_WAIT | MQGMO_FAIL_IF_QUIESCING | MQGMO_SYNCPOINT;

			/* reset the msgid and correlid */
			memcpy(msgdesc.MsgId, MQMI_NONE, sizeof(msgdesc.MsgId));
			memcpy(msgdesc.CorrelId, MQCI_NONE, sizeof(msgdesc.CorrelId));

			/* perform the MQGET */
			MQGET(qm, q, &msgdesc, &mqgmo, MAX_MESSAGE_LENGTH, msgdata, &datalen, &compcode, &reason);
			
			/* check if we got a message */
			if (MQCC_OK == compcode)
			{
				uow++;
				drainCount++;

				if (uow > batchSize)
				{
					MQCMIT(qm, &compcode, &reason);
					uow = 0;
				}
			}
		}

		if (uow > 0)
		{
			MQCMIT(qm, &compcode, &reason);
			uow = 0;
		}

		if (drainCount > 0)
		{
			printf("%d messages drained from Q prior to measurement start\n", drainCount);
		}
	}

	/* tell what we are doing */
	printf("Reading %d messages from queue %s qm %s with max wait time of %d seconds\n",
		   maxmsgs, qname, qmname, maxtime);

	/* enter get message loop */
	compcode = MQCC_OK;
	uow = 0;
	while ((MQCC_OK == compcode) && (totcount < maxmsgs))
	{
		/* set the get message options */
		mqgmo.Options = MQGMO_WAIT | MQGMO_FAIL_IF_QUIESCING | MQGMO_SYNCPOINT;
		mqgmo.WaitInterval = maxtime * 1000;
		mqgmo.MatchOptions = MQGMO_NONE;

		/* reset the msgid and correlid */
		memcpy(msgdesc.MsgId, MQMI_NONE, sizeof(msgdesc.MsgId));
		memcpy(msgdesc.CorrelId, MQCI_NONE, sizeof(msgdesc.CorrelId));

		/* perform the MQGET */
		MQGET(qm, q, &msgdesc, &mqgmo, MAX_MESSAGE_LENGTH, msgdata, &datalen, &compcode, &reason);

		/* check if we got to the end of the queue */
		/* check for errors, except for no more messages in queue */
		if (2033 == reason)
		{
			printf("Program timed out before maximum number of messages were read\n");
		}
		else
		{
			checkerror((char *)"MQGET", compcode, reason, qname);
		}

		if (compcode == MQCC_OK)
		{
			/* increase the uow count */
			uow++;

			/* check if an acknowledgement is required */
			report = msgdesc.Report;
			if ((((report & MQRO_PAN) > 0) && (fileDataPAN != NULL)) ||
				(((report & MQRO_NAN) > 0) && (fileDataNAN != NULL)))
			{
				/* either NAN or PAN is set - therefore, we need to reply */
				/* get the reply to Q and QM */
				memcpy(replyQname, msgdesc.ReplyToQ, MQ_Q_NAME_LENGTH);
				memcpy(replyQMname, msgdesc.ReplyToQMgr, MQ_Q_MGR_NAME_LENGTH);

				/* set the open options */
				strncpy(replyObjdesc.ObjectName, replyQname, MQ_Q_NAME_LENGTH);
				strncpy(replyObjdesc.ObjectQMgrName, replyQMname, MQ_Q_MGR_NAME_LENGTH);
				openopt = MQOO_OUTPUT | MQOO_FAIL_IF_QUIESCING;

				/* open the reply to queue to */
				MQOPEN(qm, &replyObjdesc, openopt, &replyq, &compcode, &reason);
				checkerror((char *)"MQOPEN(replyQ)", compcode, reason, replyQname);

				if (MQCC_OK == compcode)
				{
					/* set the fields in the MQ object descriptor for the reply */
/*					memcpy(replyObjdesc.ObjectName, replyQname, sizeof(replyObjdesc.ObjectName));
					memcpy(replyObjdesc.ObjectQMgrName, replyQMname, sizeof(replyObjdesc.ObjectQMgrName));*/

					/* set the fields in the MQMD for the reply message */
					replyMsgdesc.MsgType = MQMT_REPORT;
					memcpy(replyMsgdesc.CorrelId, msgdesc.MsgId, sizeof(replyMsgdesc.CorrelId));

					/* perform the put operation in a syncpoint */
					mqpmo.Options = MQPMO_SYNCPOINT | MQPMO_NEW_MSG_ID;

					/* write the report message */
					if (((report & MQRO_PAN) > 0) && (fileDataPAN != NULL))
					{
						/* check if the data has an RFH */
						if ((fileSizePAN > 32) && (memcmp(fileDataPAN, "RFH ", 4) == 0))
						{
							memcpy(replyMsgdesc.Format, "MQHRF2  ", 8);
						}

						MQPUT(qm, replyq, &replyMsgdesc, &mqpmo, fileSizePAN, fileDataPAN, &compcode, &reason);
					}
					else
					{
						/* check if the data has an RFH */
						if ((fileSizeNAN > 32) && (memcmp(fileDataNAN, "RFH ", 4) == 0))
						{
							memcpy(replyMsgdesc.Format, "MQHRF2  ", 8);
						}

						MQPUT(qm, replyq, &replyMsgdesc, &mqpmo, fileSizeNAN, fileDataNAN, &compcode, &reason);
					}

					checkerror((char *)"MQOPEN(reply)", compcode, reason, replyQname);

					/* force a syncpoint now */
					MQCMIT(qm, &compcode, &reason);
					checkerror((char *)"MQCMIT(reply)", compcode, reason, qmname);
					uow = 0;

					/* close the queue */
					MQCLOSE(qm, &replyq, MQCO_NONE , &compcode, &reason);
					checkerror((char *)"MQCLOSE(reply)", compcode, reason, replyQname);
				}
			}

			/* check if we are at the maximum batch size */
			if (uow > batchSize)
			{
				MQCMIT(qm, &compcode, &reason);
				checkerror((char *)"MQCMIT", compcode, reason, qmname);
				uow = 0;
			}

			/* only do this step if the MQGET worked */
			/* get the message time from the MQMD */
			memcpy(currtime, msgdesc.PutTime, 8);

			/* check if the time is the same or not */
			/* only check down to the seconds position */
			/* if ((prevtime[0] == 0) || (memcmp(prevtime, currtime, 6) == 0))*/
			currtime_secs = ( atol(currtime) / 100 );
			if ((0 == prevtime_secs) || (currtime_secs <= prevtime_secs ))
			{
				msgcount++;
				totcount++;
			}
			else
			{
				/* get the average of the last 10 intervals */
				recent10 = 0;
				recent10sec = last10secs[9];
				for (i=9; i>0; i--)
				{
					last10secs[i] = last10secs[i - 1];
					last10[i] = last10[i - 1];
					recent10 += last10[i - 1];
				}

				/* get the most recent count */
				last10[0] = msgcount;
				last10secs[0] = currtime_secs;
				recent10 += msgcount;

				/* write out the number of messages in this second */
				sprintf(msgPtr,"%6.6s %d msgs", prevtime, msgcount);
				msgPtr += strlen(msgPtr);

				/* calculate the recent average */
				if (recent10sec > 0)
				{
					/* convert the time from hhmmss format to pure seconds */
					ss = currtime_secs % 100;
					mm = ((currtime_secs - ss) % 10000) / 100;
					hh = (currtime_secs - mm - ss) / 10000;
					time1sec = (hh * 3600) + (mm * 60) + ss;

					ss = recent10sec % 100;
					mm = ((recent10sec - ss) % 10000) / 100;
					hh = (recent10sec - mm - ss) / 10000;
					time2sec = (hh * 3600) + (mm * 60) + ss;

					avgrate = (recent10) / (time1sec - time2sec);

					/* write out the number of messages in this second */
					sprintf(msgPtr, " - recent average = %7.2f\n", avgrate);
				}
				else
				{

					/* no average yet - just start a new line */
					msgPtr[0] = '\n';
					msgPtr++;
				}

				msgPtr += strlen(msgPtr);
				reportCount++;
				if (reportCount >= reportInterval)
				{
					printf("%s", msgArea);
					reportCount = 0;
					msgPtr = msgArea;
					msgArea[0] = 0;
				}

				/* is this the first time through? */
				if (firstsec == 0)
				{
					/* remember count in first second */
					firstsec = msgcount;
				}

				/* keep track of the maximum message rate */
				if (msgcount > maxrate)
				{
					maxrate = msgcount;
				}

				/* reset the messages in second counter */
				msgcount = 1;

				/* count the total number of messages */
				totcount++;

				/* count the number of individual seconds with at least one message */
				secondcount++;
			}

			if (currtime_secs > prevtime_secs )
			{
				memcpy(prevtime, currtime, 8);
				prevtime_secs = currtime_secs;
			}

			/* calculate the total bytes in the message */
			totalbytes += datalen;
		}
	}

	/* count the last interval */
	secondcount++;

	/* check if we have a uow open */
	if (uow > 0)
	{
		MQCMIT(qm, &compcode, &reason);
		checkerror((char *)"MQCMIT", compcode, reason, qmname);
	}

	/* dump out the last time interval */
	sprintf(msgPtr, "%6.6s %d\n", prevtime, msgcount);
	printf("%s", msgArea);

	/* keep track of the maximum message rate */
	if (msgcount > maxrate)
	{
		maxrate = msgcount;
	}

	/* close the input queue */
	printf("closing the input queue (%s)\n", qname);
	MQCLOSE(qm, &q, MQCO_NONE, &compcode, &reason);

	checkerror((char *)"MQCLOSE", compcode, reason, qname);

	/* Disconnect from the queue manager */
	printf("disconnecting from the queue manager\n");
	MQDISC(&qm, &compcode, &reason);

	checkerror((char *)"MQDISC", compcode, reason, qmname);

	/* dump out the total message count */
	printf("Total messages %d\n", totcount);

	/* give the total and average message size */
	if (totcount > 0)
	{
		avgbytes = totalbytes / totcount;
		printf("total bytes in all messages %d\n", totalbytes);
		printf("average message size %d\n", avgbytes);
	}

	/* indicate the number of seconds with at least one message */
	printf("Total number of seconds with at least one message %d\n", secondcount);

	/* write out the average message rate, ignoring the first and last intervals */
	if (secondcount > 2)
	{
		tempnumb[0] = 0;
		sprintf(tempnumb, "%d", totcount - firstsec - msgcount);
		fCount = atof(tempnumb);
		avgrate = (fCount) / (secondcount - 2);
		printf("Average message rate except first and last intervals %7.2f\n", avgrate);
	}

	/* print out the maximum rate */
	printf("Peak message rate %d\n", maxrate);

	/******************************************************************/
	/*                                                                */
	/* END OF PROGRAM                                                 */
	/*                                                                */
	/******************************************************************/
	printf("MQTIMES2 program ended\n");

	if (fileDataPAN != NULL)
	{
		free(fileDataPAN);
	}

	if (fileDataNAN != NULL)
	{
		free(fileDataNAN);
	}

	return 0;
}
